@@ -10,6 +10,7 @@
    compute relocation and export tables, rewrite some COFF files,
    call the native linker *)
 
+open Compat
 open Coff
 open Cmdline
 
@@ -96,11 +97,14 @@ let get_output ?(use_bash = false) ?(accept_error=false) cmd =
       Sys.remove fn;
       r
 
-let get_output1 ?use_bash cmd =
-  match get_output ?use_bash cmd with
-  | output::_ -> output
-  | [] -> raise (Failure ("command " ^ cmd ^ " did not return any output"))
+let get_output1 ?use_bash ?accept_error fmt =
+  Printf.ksprintf (fun cmd ->
+    match get_output ?use_bash ?accept_error cmd with
+    | output::_ -> output
+    | [] -> raise (Failure ("command " ^ cmd ^ " did not return any output"))) fmt
 
+let get_output ?use_bash ?accept_error fmt =
+  Printf.ksprintf (get_output ?use_bash ?accept_error) fmt
 
 (* Preparing command line *)
 
@@ -195,7 +199,7 @@ type cmdline = {
 let new_cmdline () =
   let rf = match !toolchain with
   | `MSVC | `MSVC64 | `LIGHTLD -> true
-  | `MINGW | `MINGW64 | `GNAT | `GNAT64 | `CYGWIN | `CYGWIN64 -> false
+  | `MINGW | `MINGW64 | `GNAT | `GNAT64 | `CYGWIN64 -> false
   in
   {
    may_use_response_file = rf;
@@ -207,7 +211,7 @@ let run_command cmdline cmd =
   in
   (* note: for Cygwin, using bash allow to follow symlinks to find
      gcc... *)
-  if !toolchain = `CYGWIN || !toolchain = `CYGWIN64 ||
+  if !toolchain = `CYGWIN64 ||
      String.length cmd + String.length silencer > max_command_length
   then begin
     (* Dump the command in a text file and apply bash to it. *)
@@ -244,18 +248,40 @@ let quote_files cmdline lst =
 
 (* Looking for files *)
 
-let cygpath l =
-  get_output (Printf.sprintf "cygpath -m %s" (String.concat " " (List.map Filename.quote l)))
+let cygpath l cont =
+  let accept_error = (!use_cygpath = `Try && l <> []) in
+  let l =
+    let args = String.concat " " (List.map Filename.quote l) in
+    get_output ~accept_error "cygpath -m %s" args
+  in
+  if accept_error && l = [] then begin
+    use_cygpath := `No;
+    None
+  end else
+    cont l
 
 let cygpath1 fn =
-  get_output1 (Printf.sprintf "cygpath -m %s" fn)
+  let accept_error = (!use_cygpath = `Try) in
+  match get_output ~accept_error "cygpath -m %s" fn with
+  | output::_ ->
+      if accept_error then
+        use_cygpath := `Yes;
+      Some output
+  | [] when accept_error ->
+      use_cygpath := `No;
+      None
+  | [] ->
+      raise (Failure "cygpath did not return any output")
 
 let file_exists fn =
   if Sys.file_exists fn && not (Sys.is_directory fn) then Some fn
-  else if !use_cygpath && Sys.file_exists (fn ^ ".lnk") then
-    Some (cygpath1 fn)
+  else if !use_cygpath <> `No && Sys.file_exists (fn ^ ".lnk") then
+    cygpath1 fn
   else None
 
+let dir_exists_no_cygpath fn =
+  Sys.file_exists fn && (try Sys.is_directory fn with Sys_error _ -> false)
+
 let rec find_file_in = function
   | [] -> None
   | fn::rest ->
@@ -263,17 +289,20 @@ let rec find_file_in = function
         | Some x -> Some x
         | None -> find_file_in rest
 
-let find_file fn =
+let find_file suffixes fn =
   let l =
     List.flatten
       (List.map
          (fun dir ->
             let fn = Filename.concat dir fn in
-            [ fn; fn ^ ".lib"; fn ^ ".dll.a"; fn ^ ".a" ]
+            fn :: (List.map (fun suff -> fn ^ suff) suffixes)
          ) (""::!search_path)) in
   match find_file_in l with
     | Some x -> Some x
-    | None -> if !use_cygpath then find_file_in (cygpath l) else None
+    | None ->
+        if !use_cygpath <> `No then
+          cygpath l find_file_in
+        else None
 
 let rec map_until_found f = function
   | [] ->
@@ -293,16 +322,22 @@ let find_file =
     with Not_found ->
       try Hashtbl.find memo (k ^ ".lib")
       with Not_found ->
-        let fns =
+        let fns, suffixes =
+          (* XXX Not sure why we do these extensions for _both_ MSVC and
+                 mingw-w64 rather than .lib for MSVC and the .a ones for
+                 mingw-w64? *)
+          let standard_suffixes = [".lib"; ".dll.a"; ".a"] in
           if String.length fn > 2 && String.sub fn 0 2 = "-l" then
             let base = String.sub fn 2 (String.length fn - 2) in
-            if !toolchain = `MSVC || !toolchain = `MSVC64 then
-              ["lib" ^ base; base]
+            if String.length base > 0 && base.[0] = ':' then
+              [String.sub base 1 (String.length base - 1)], []
+            else if !toolchain = `MSVC || !toolchain = `MSVC64 then
+              ["lib" ^ base; base], standard_suffixes
             else
-              ["lib" ^ base]
-          else [fn] in
+              ["lib" ^ base], standard_suffixes
+          else [fn], standard_suffixes in
         let r =
-          match map_until_found find_file fns with
+          match map_until_found (find_file suffixes) fns with
           | Some fn -> fn
           | None ->
               failwith (Printf.sprintf "Cannot find file %S" fn)
@@ -602,7 +637,7 @@ let parse_dll_exports fn =
 let dll_exports fn = match !toolchain with
   | `MSVC | `MSVC64 | `LIGHTLD ->
       failwith "Creation of import library not supported for this toolchain"
-  | `GNAT | `GNAT64 | `CYGWIN | `CYGWIN64 | `MINGW | `MINGW64 ->
+  | `GNAT | `GNAT64 | `CYGWIN64 | `MINGW | `MINGW64 ->
       let dmp = temp_file "dyndll" ".dmp" in
       if cmd_verbose (Printf.sprintf "%s -p %s > %s" !objdump fn dmp) <> 0
       then failwith "Error while extracting exports from a DLL";
@@ -616,10 +651,6 @@ let patch_output filename =
         if not (Sys.file_exists filename) && (Sys.file_exists (filename ^ ".exe")) then filename ^ ".exe"
         else filename
       in
-      let filename =
-        if !use_cygpath then cygpath1 filename
-        else filename
-      in
       begin try Stacksize.set_stack_reserve filename x
       with exn ->
         Printf.eprintf "Cannot set stack reserve: %s\n%!"
@@ -634,12 +665,19 @@ let patch_output filename =
    Indeed, we will create "__imp_X" (with a redirection to "X").
    Collect such cases in "imported".
 *)
-let needed imported defined unalias obj =
+let needed imported defined resolve_alias resolve_alternate obj =
   let rec normalize name =
     try
-      let r = unalias name in
+      let r = resolve_alias name in
       if r <> name then normalize r else r
-    with Not_found -> name
+    with Not_found ->
+      (* Fall back to alternate name if and only if name was not found.
+         https://devblogs.microsoft.com/oldnewthing/20200731-00/?p=104024 *)
+      try
+        let r = resolve_alternate name in
+        if r <> name then normalize r else r
+      with Not_found ->
+        name
   in
   let normalize_imp name =
     match check_prefix "__imp_" name with
@@ -701,8 +739,9 @@ let build_dll link_exe output_file files exts extra_args =
   in
   (* Collect all the available symbols, including those defined
      in default libraries *)
-  let defined, from_imports, unalias =
+  let defined, from_imports, resolve_alias, resolve_alternate =
     let aliases = Hashtbl.create 16 in
+    let alternates = Hashtbl.create 16 in
     let defined = ref StrSet.empty in
     let from_imports = ref StrSet.empty in (* symbols from import libraries *)
     let add_def s = defined := StrSet.add s !defined in
@@ -728,6 +767,21 @@ let build_dll link_exe output_file files exts extra_args =
         )
         (Coff.aliases obj);
 
+      (* Collect alternatenames *)
+      let collect_alternatenames alternatenames =
+        List.iter (fun s -> match String.split_on_char '=' s with
+            | [alternate_name; true_name] ->
+                debug 2 "alternatename %s -> %s" alternate_name true_name;
+                Hashtbl.add alternates alternate_name true_name
+            | _ ->
+                debug 1 "alternatenames unrecognized: %s" s;
+          ) alternatenames
+      in
+      List.iter (function
+          | ("alternatename", alternatenames) ->
+              collect_alternatenames alternatenames
+          | _ -> ())
+        (Coff.directives obj);
 
       (* Iterates through DEFAULTLIB directives *)
       let register_deflib fn =
@@ -787,7 +841,7 @@ let build_dll link_exe output_file files exts extra_args =
     if !machine = `x64 then add_def "__ImageBase"
     else add_def "___ImageBase";
 
-    !defined, !from_imports, (Hashtbl.find aliases)
+    !defined, !from_imports, (Hashtbl.find aliases), (Hashtbl.find alternates)
   in
 
   (* Determine which objects from the given libraries should be linked
@@ -824,7 +878,7 @@ let build_dll link_exe output_file files exts extra_args =
   let imported = ref StrSet.empty in
 
   let imports obj =
-    let n = needed imported defined unalias obj in
+    let n = needed imported defined resolve_alias resolve_alternate obj in
     imported_from_implib := StrSet.union !imported_from_implib (StrSet.inter n from_imports);
     let undefs = StrSet.diff n defined in
     StrSet.filter
@@ -904,7 +958,7 @@ let build_dll link_exe output_file files exts extra_args =
             if !explain then
               Printf.printf "%s needs %s (not found)\n%!" fn s
       )
-      (needed imported defined unalias obj)
+      (needed imported defined resolve_alias resolve_alternate obj)
 
   and link_libobj (libname,objname,obj) =
     if Hashtbl.mem libobjects (libname,objname) then ()
@@ -1083,7 +1137,7 @@ let build_dll link_exe output_file files exts extra_args =
           !subsystem
           files descr
           extra_args
-    | `CYGWIN | `CYGWIN64 ->
+    | `CYGWIN64 ->
         let def_file =
           if main_pgm then ""
           else
@@ -1190,7 +1244,7 @@ let read_gnatls () =
    (* This function is used by the GNAT toolchain to compute the include
       directory. gnatls actually returns with an error code different to 0, so
       we need to accept the error here. *)
-   let str_l = get_output ~accept_error:true ("gnatls -v") in
+   let str_l = get_output ~accept_error:true "gnatls -v" in
    let ada_include =
      List.hd (List.filter (fun s -> ends_with s "adainclude") str_l) in
    Filename.dirname (strip ada_include)
@@ -1213,6 +1267,13 @@ let nsplit str sep =
     loop [] 0
 
 let normalize_path path =
+  let path =
+    if Sys.win32 then
+      let back_to_forward c = if c = '\\' then '/' else c in
+      String.init (String.length path) (fun i -> back_to_forward path.[i])
+    else
+      path
+  in
   let path = nsplit path '/' in
   let rec loop acc path =
     match path with
@@ -1226,7 +1287,7 @@ let normalize_path path =
     | [] -> List.rev acc
   in
   let path = loop [] path in
-  String.concat "/" path
+  String.concat Filename.dir_sep path
 
 let remove_duplicate_paths paths =
   let set = Hashtbl.create 16 in
@@ -1248,21 +1309,45 @@ let setup_toolchain () =
   let mingw_libs pre =
     gcc := pre ^ "gcc";
     objdump := pre ^ "objdump";
-    let rec get_lib_search_dirs input =
+    let rec get_lib_search_dirs install libraries input =
       match input with
       | entry :: input ->
-          begin try
+          if String.length entry > 9 && String.sub entry 0 9 = "install: " then
+            get_lib_search_dirs (String.sub entry 9 (String.length entry - 9)) libraries input
+          else begin try
             match split entry '=' with
-            | "libraries: ", paths -> nsplit paths ':'
-            | _ -> get_lib_search_dirs input
+            | "libraries: ", paths -> get_lib_search_dirs install paths input
+            | _ -> get_lib_search_dirs install libraries input
           with Not_found ->
-            get_lib_search_dirs input
+            get_lib_search_dirs install libraries input
           end
-      | [] -> []
+      | [] ->
+          let install =
+            if install <> "" then
+              (* Ensure install does not end with a separator (or
+                 Sys.is_directory will fail) *)
+              Filename.concat install Filename.current_dir_name
+              |> Filename.dirname
+            else
+              ""
+          in
+          let separator, run_through_cygpath =
+            if Sys.win32 then
+              if dir_exists_no_cygpath install then
+                ';', false
+              else
+                ':', (!use_cygpath <> `No)
+            else
+              ':', false
+          in
+          let libraries = nsplit libraries separator in
+          if run_through_cygpath then
+            Option.value ~default:libraries (cygpath libraries Option.some)
+          else
+            libraries
     in
     let lib_search_dirs =
-      get_output (!gcc ^ " -print-search-dirs")
-      |> get_lib_search_dirs
+      get_lib_search_dirs "" "" (get_output "%s -print-search-dirs" !gcc)
       |> List.map normalize_path
       |> remove_duplicate_paths
     in
@@ -1283,7 +1368,7 @@ let setup_toolchain () =
       search_path := !dirs;
       add_flexdll_obj := false;
       noentry := true
-  | `CYGWIN | `CYGWIN64 ->
+  | `CYGWIN64 ->
       gcc := "gcc";
       objdump := "objdump";
       search_path :=
@@ -1294,7 +1379,7 @@ let setup_toolchain () =
            Filename.dirname (get_output1 ~use_bash:true "gcc -print-libgcc-file-name");
           ];
       default_libs := ["-lkernel32"; "-luser32"; "-ladvapi32";
-                       "-lshell32"; "-lcygwin"; "-lgcc"]
+                       "-lshell32"; "-lcygwin"; "-lgcc_s"; "-lgcc"]
   | `MSVC | `MSVC64 ->
       search_path := !dirs @
         parse_libpath (try Sys.getenv "LIB" with Not_found -> "");
@@ -1310,7 +1395,7 @@ let setup_toolchain () =
     search_path :=
       !dirs @
       [
-       Filename.dirname (get_output1 (!gcc ^ " -print-libgcc-file-name"));
+       Filename.dirname (get_output1 "%s -print-libgcc-file-name" !gcc);
        read_gnatls ();
       ];
     default_libs :=
@@ -1354,7 +1439,7 @@ let compile_if_needed file =
             (mk_dirs_opt "/I")
             file
             pipe
-      | `CYGWIN | `CYGWIN64 ->
+      | `CYGWIN64 ->
           Printf.sprintf
             "gcc -c -o %s %s %s"
             (Filename.quote tmp_obj)
@@ -1408,7 +1493,6 @@ let all_files () =
   let tc = match !toolchain with
   | `MSVC -> "msvc.obj"
   | `MSVC64 -> "msvc64.obj"
-  | `CYGWIN -> "cygwin.o"
   | `CYGWIN64 -> "cygwin64.o"
   | `MINGW64 -> "mingw64.o"
   | `GNAT -> "gnat.o"
@@ -1425,25 +1509,14 @@ let main () =
   parse_cmdline ();
   setup_toolchain ();
 
-  use_cygpath :=
-    begin
-      match !toolchain, !cygpath_arg with
-      | _, `Yes -> true
-      | _, `No -> false
-      | (`GNAT|`GNAT64|`MINGW|`MINGW64|`CYGWIN|`CYGWIN64), `None ->
-          begin match Sys.os_type with
-          | "Unix" | "Cygwin" ->
-              Sys.command "cygpath -S 2>/dev/null >/dev/null" = 0
-          | "Win32" ->
-              Sys.command "cygpath -S 2>NUL >NUL" = 0
-          | _ -> assert false
-          end
-      | (`MSVC|`MSVC64|`LIGHTLD), `None -> false
-    end;
-
-
   if !verbose >= 2 then (
-    Printf.printf "** Use cygpath: %b\n" !use_cygpath;
+    let use_cygpath =
+      match !use_cygpath with
+      | `Yes -> "true"
+      | `No -> "false"
+      | `Try -> "maybe"
+    in
+    Printf.printf "** Use cygpath: %s\n" use_cygpath;
     Printf.printf "** Search path:\n";
     List.iter print_endline !search_path;
     if !use_default_libs then begin
@@ -1459,7 +1532,12 @@ let main () =
       build_dll !exe_mode !output_file files !exts
         (String.concat " " (List.map Filename.quote (List.rev !extra_args)))
   | `PATCH ->
-      patch_output !output_file
+      let output_file =
+        if !use_cygpath <> `No then
+          Option.value ~default:!output_file (cygpath1 !output_file)
+        else
+          !output_file in
+      patch_output output_file
 
 let () =
   try main ()
