@@ -9,10 +9,9 @@
 (*                                                                        *)
 (**************************************************************************)
 
-open OpamCompat
-
 type install_warning =
-  [ `Add_exe | `Install_dll | `Install_script | `Install_unknown | `Cygwin | `Cygwin_libraries ]
+  [ `Add_exe | `Install_dll | `Install_script | `Install_unknown
+  | `Cygwin | `Msys2 | `Tainted of [`Msys2 | `Cygwin] | `Cygwin_libraries ]
 type install_warning_fn = string -> install_warning -> unit
 
 exception Process_error of OpamProcess.result
@@ -56,15 +55,45 @@ let file_or_symlink_exists f =
 
 let (/) = Filename.concat
 
+let real_path p =
+  (* if Filename.is_relative p then *)
+    match (try Some (Sys.is_directory p) with Sys_error _ -> None) with
+    | None ->
+      let rec resolve dir =
+        if Sys.file_exists dir then OpamCompat.Unix.realpath dir else
+        let parent = Filename.dirname dir in
+        if dir = parent then dir
+        else Filename.concat (resolve parent) (Filename.basename dir)
+      in
+      let p =
+        if Filename.is_relative p then Filename.concat (Sys.getcwd ()) p
+        else p
+      in
+      resolve p
+    | Some true -> OpamCompat.Unix.realpath p
+    | Some false ->
+      let dir = OpamCompat.Unix.realpath (Filename.dirname p) in
+      match Filename.basename p with
+      | "." -> dir
+      | base -> dir / base
+  (* else p *)
+
 let temp_basename prefix =
   Printf.sprintf "%s-%d-%06x" prefix (OpamStubs.getpid ()) (Random.int 0xFFFFFF)
 
 let rec mk_temp_dir ?(prefix="opam") () =
   let s = Filename.get_temp_dir_name () / temp_basename prefix in
   if Sys.file_exists s then
-    mk_temp_dir ()
+    mk_temp_dir ~prefix ()
   else
-    s
+    real_path s
+
+let rec mk_unique_dir ~dir ?(prefix="opam") () =
+  let s = dir / Printf.sprintf "%s-%06x" prefix (Random.int 0xFFFFFF) in
+  if Sys.file_exists s then
+    mk_unique_dir ~dir ~prefix ()
+  else
+    real_path s
 
 let safe_mkdir dir =
   try
@@ -81,19 +110,66 @@ let mkdir dir =
     end in
   aux dir
 
-let rm_command =
-  if Sys.win32 then
-    "cmd /d /v:off /c rd /s /q"
-  else
-    "rm -rf"
+let get_files dirname =
+  let dir = Unix.opendir dirname in
+  let rec aux files =
+    match Unix.readdir dir with
+    | "." | ".." -> aux files
+    | file -> aux (file :: files)
+    | exception End_of_file -> files
+  in
+  let files = aux [] in
+  Unix.closedir dir;
+  files
+
+let log_for_file_management () =
+  OpamCoreConfig.(!r.debug_level) >= 4
+
+(* From stdune/src/fpath.ml *)
+let win32_unlink fn =
+  try Unix.unlink fn
+  with Unix.Unix_error (Unix.EACCES, _, _) as e -> (
+    try
+      (* Try removing the read-only attribute *)
+      Unix.chmod fn 0o666;
+      Unix.unlink fn
+    with _ -> raise e)
+
+let remove_file_t ?(with_log=true) file =
+  try
+    if with_log || log_for_file_management () then
+      log "rm %s" file;
+    if Sys.win32 then
+      win32_unlink file
+    else
+      Unix.unlink file
+  with Unix.Unix_error _ as e ->
+    internal_error "Cannot remove %s (%s)." file (Printexc.to_string e)
+
+let rec remove_dir_t dir =
+  let files = get_files dir in
+  List.iter (fun file ->
+      let file = Filename.concat dir file in
+      match Unix.lstat file with
+      | {Unix.st_kind = Unix.S_DIR; _} ->
+        remove_dir_t file
+      | {Unix.st_kind = Unix.(S_REG | S_LNK | S_CHR | S_BLK | S_FIFO | S_SOCK); _} ->
+        remove_file_t ~with_log:false file
+      | exception Unix.(Unix_error (ENOENT, _, _)) when Sys.win32 ->
+        (* This is usually caused by invalid symlinks extracted by a
+           Cygwin/MSYS2 tar. *)
+        remove_file_t ~with_log:false file
+    ) files;
+  Unix.rmdir dir
 
 let remove_dir dir =
   log "rmdir %s" dir;
-  if Sys.file_exists dir then (
-    let err = Sys.command (Printf.sprintf "%s %s" rm_command (Filename.quote dir)) in
-      if err <> 0 then
-        internal_error "Cannot remove %s (error %d)." dir err
-  )
+  if Sys.file_exists dir then begin
+    if Sys.is_directory dir then
+      remove_dir_t dir
+    else
+      remove_file_t ~with_log:true dir
+  end
 
 let temp_files = Hashtbl.create 1024
 let logs_cleaner =
@@ -134,9 +210,13 @@ let remove_file file =
   if
     try ignore (Unix.lstat file); true with Unix.Unix_error _ -> false
   then (
+    log "rm %s" file;
     try
-      log "rm %s" file;
-      Unix.unlink file
+      try Unix.unlink file
+      with Unix.Unix_error(EACCES, _, _) when Sys.win32 ->
+        (* Attempt to remove the read-only bit on Windows *)
+        Unix.chmod file 0o666;
+        Unix.unlink file
     with Unix.Unix_error _ as e ->
       internal_error "Cannot remove %s (%s)." file (Printexc.to_string e)
   )
@@ -186,18 +266,23 @@ let setup_copy ?(chmod = fun x -> x) ~src ~dst () =
             remove_file dst
       with Unix.Unix_error(ENOENT, _, _) -> ()
     in
-    let oc =
-      open_out_gen
-        [ Open_wronly; Open_creat; Open_trunc; Open_binary ]
-        perm dst
+    let fd =
+      let flags = Unix.[ O_WRONLY; O_CREAT; O_TRUNC ] in
+      try Unix.openfile dst flags perm
+      with Unix.Unix_error(EACCES, _, _) when Sys.win32 ->
+        (* Attempt to remove the read-only bit on Windows *)
+        begin
+          try Unix.chmod dst 0o666
+          with Unix.Unix_error(_, _, _) -> ()
+        end;
+        Unix.openfile dst flags perm
     in
-    let fd = Unix.descr_of_out_channel oc in
     try
       if Unix.((fstat fd).st_perm) <> perm then
         Unix.fchmod fd perm;
-      (ic, oc)
+      (ic, Unix.out_channel_of_descr fd)
     with exn ->
-      OpamStd.Exn.finalise exn (fun () -> close_out oc)
+      OpamStd.Exn.finalise exn (fun () -> Unix.close fd)
   with exn ->
     OpamStd.Exn.finalise exn (fun () -> close_in ic)
 
@@ -304,6 +389,10 @@ let with_tmp_dir fn =
     OpamStd.Exn.finalise e @@ fun () ->
     remove_dir dir
 
+let in_tmp_dir fn =
+  with_tmp_dir @@ fun dir ->
+    in_dir dir fn
+
 let with_tmp_dir_job fjob =
   let dir = mk_temp_dir () in
   mkdir dir;
@@ -315,62 +404,8 @@ let remove file =
   else
     remove_file file
 
-(* Sets path to s and returns the old path *)
-let getchdir s =
-  let p =
-    try Sys.getcwd ()
-    with Sys_error _ ->
-      let p = OpamCoreConfig.(!r.log_dir) in
-      mkdir p; p
-  in
-  chdir s;
-  p
-
-let normalize s =
-  try getchdir (getchdir s) with File_not_found _ -> s
-
-let real_path p =
-  (* if Filename.is_relative p then *)
-    match (try Some (Sys.is_directory p) with Sys_error _ -> None) with
-    | None ->
-      let rec resolve dir =
-        if Sys.file_exists dir then normalize dir else
-        let parent = Filename.dirname dir in
-        if dir = parent then dir
-        else Filename.concat (resolve parent) (Filename.basename dir)
-      in
-      let p =
-        if Filename.is_relative p then Filename.concat (Sys.getcwd ()) p
-        else p
-      in
-      resolve p
-    | Some true -> normalize p
-    | Some false ->
-      let dir = normalize (Filename.dirname p) in
-      match Filename.basename p with
-      | "." -> dir
-      | base -> dir / base
-  (* else p *)
-
 type command = string list
 
-let default_env =
-  Unix.environment ()
-
-let env_var env var =
-  let len = Array.length env in
-  let f = if Sys.win32 then String.uppercase_ascii else fun x -> x in
-  let prefix = f var^"=" in
-  let pfxlen = String.length prefix in
-  let rec aux i =
-    if i >= len then "" else
-    let s = env.(i) in
-    if OpamStd.String.starts_with ~prefix (f s) then
-      String.sub s pfxlen (String.length s - pfxlen)
-    else aux (i+1)
-  in
-  aux 0
-
 let forward_to_back =
   if Sys.win32 then
     String.map (function '/' -> '\\' | c -> c)
@@ -383,109 +418,73 @@ let back_to_forward =
   else
     fun x -> x
 
-(* OCaml 4.05.0 no longer follows the updated PATH to resolve commands. This
-   makes unqualified commands absolute as a workaround. *)
-let t_resolve_command =
-  let is_external_cmd name =
-    let name = forward_to_back name in
-    OpamStd.String.contains_char name Filename.dir_sep.[0]
-  in
-  let check_perms =
-    if Sys.win32 then fun f ->
-      try (Unix.stat f).Unix.st_kind = Unix.S_REG
-      with e -> OpamStd.Exn.fatal e; false
-    else fun f ->
-      try
-        let open Unix in
-        let uid = geteuid () in
-        let groups = OpamStd.IntSet.of_list (getegid () :: Array.to_list (getgroups ())) in
-        let {st_uid; st_gid; st_perm; _} = stat f in
-        let mask =
-          if uid = st_uid then
-            0o100
-          else if OpamStd.IntSet.mem st_gid groups then
-            0o010
-          else
-            0o001
-        in
-        if (st_perm land mask) <> 0 then
-          true
-        else
-          match OpamACL.get_acl_executable_info f st_uid with
-          | None -> false
-          | Some [] -> true
-          | Some gids -> OpamStd.IntSet.(not (is_empty (inter (of_list gids) groups)))
-      with e -> OpamStd.Exn.fatal e; false
-  in
-  let resolve ?dir env name =
-    if not (Filename.is_relative name) then begin
-      (* absolute path *)
-      if not (Sys.file_exists name) then `Not_found
-      else if not (check_perms name) then `Denied
-      else `Cmd name
-    end else if is_external_cmd name then begin
-      (* relative path *)
-      let cmd = match dir with
-        | None -> name
-        | Some d -> Filename.concat d name
-      in
-      if not (Sys.file_exists cmd) then `Not_found
-      else if not (check_perms cmd) then `Denied
-      else `Cmd cmd
-    end else
-    (* bare command, lookup in PATH *)
-    (* Following the shell sematics for looking up PATH, programs with the
-       expected name but not the right permissions are skipped silently.
-       Therefore, only two outcomes are possible in that case, [`Cmd ..] or
-       [`Not_found]. *)
-    let path = OpamStd.Sys.split_path_variable (env_var env "PATH") in
-    let name =
-      if Sys.win32 && not (Filename.check_suffix name ".exe") then
-        name ^ ".exe"
-      else name
-    in
-    let possibles = OpamStd.List.filter_map (fun path ->
-        let candidate = Filename.concat path name in
-        if Sys.file_exists candidate then Some candidate else None) path
-    in
-    match List.find check_perms possibles with
-    | cmdname -> `Cmd cmdname
-    | exception Not_found ->
-      if possibles = [] then
-        `Not_found
-      else
-        `Denied
-  in
-  fun ?(env=default_env) ?dir name ->
-    resolve env ?dir name
+let resolve_command = OpamProcess.resolve_command
 
-let resolve_command ?env ?dir name =
-  match t_resolve_command ?env ?dir name with
-  | `Cmd cmd -> Some cmd
-  | `Denied | `Not_found -> None
+let bin_contains_bash =
+  if not Sys.win32 && not Sys.cygwin then fun _ -> false else
+  fun bin ->
+    (resolve_command ~env:[||] (Filename.concat bin "bash.exe"))
+    <> None
 
 let apply_cygpath name =
-  let r =
-    OpamProcess.run
-      (OpamProcess.command ~name:(temp_file "command") ~verbose:false "cygpath" ["--"; name])
-  in
-  OpamProcess.cleanup ~force:true r;
-  if OpamProcess.is_success r then
-    List.hd r.OpamProcess.r_stdout
-  else
+  (* XXX Deeper bug, looking in the cygvoke code (see OpamProcess.create) *)
+  match resolve_command "cygpath" with
+  | Some cygpath ->
+    let r =
+      OpamProcess.run
+        (OpamProcess.command ~name:(temp_file "command")
+           ~allow_stdin:false ~verbose:false cygpath ["--"; name])
+    in
+    OpamProcess.cleanup ~force:true r;
+    if OpamProcess.is_success r then
+      match r.OpamProcess.r_stdout with
+      | l::_ -> l
+      | _ -> ""
+    else
+      OpamConsole.error_and_exit `Internal_error "Could not apply cygpath to %s" name
+  | None ->
     OpamConsole.error_and_exit `Internal_error "Could not apply cygpath to %s" name
 
 let get_cygpath_function =
   if Sys.win32 then
     fun ~command ->
-      lazy (if OpamStd.(Option.map_default Sys.is_cygwin_variant `Native (resolve_command command)) = `Cygwin then
-              apply_cygpath
-            else
-              fun x -> x)
+      lazy (
+        if OpamStd.Option.map_default
+            (OpamStd.Sys.is_cygwin_variant
+               ?search_in_first:(OpamCoreConfig.(!r.cygbin)))
+               false
+               (resolve_command command) then
+          apply_cygpath
+        else fun x -> x
+      )
   else
     let f = Lazy.from_val (fun x -> x) in
     fun ~command:_ -> f
 
+let apply_cygpath_path_transform ~pathlist cygpath path =
+  let args = if pathlist then [ "--path" ] else [] in
+  let r =
+    OpamProcess.run
+      (OpamProcess.command ~name:(temp_file "command") ~verbose:false
+         cygpath (args @ ["--"; path]))
+  in
+  OpamProcess.cleanup ~force:true r;
+  if OpamProcess.is_success r then
+    List.hd r.OpamProcess.r_stdout
+  else
+    OpamConsole.error_and_exit `Internal_error "Could not apply cygpath --path to %s" path
+
+let get_cygpath_path_transform =
+  (* We are running in a functioning Cygwin or MSYS2 environment if and only
+     if `cygpath` is in the PATH. *)
+  if Sys.win32 then
+    lazy (
+      match resolve_command "cygpath" with
+      | Some cygpath -> apply_cygpath_path_transform cygpath
+      | None -> fun ~pathlist:_ x -> x)
+  else
+    Lazy.from_val (fun ~pathlist:_ x -> x)
+
 let runs = ref []
 let print_stats () =
   match !runs with
@@ -497,18 +496,16 @@ let print_stats () =
 let log_file ?dir name = temp_file ?dir (OpamStd.Option.default "log" name)
 
 let make_command
-    ?verbose ?(env=default_env) ?name ?text ?metadata ?allow_stdin ?stdout
+    ?verbose ?env ?name ?text ?metadata ?allow_stdin ?stdout
     ?dir ?(resolve_path=true)
     cmd args =
+  let env = match env with None -> OpamProcess.default_env () | Some e -> e in
   let name = log_file name in
   let verbose =
     OpamStd.Option.default OpamCoreConfig.(!r.verbose_level >= 2) verbose
   in
-  (* Check that the command doesn't contain whitespaces *)
-  if None <> try Some (String.index cmd ' ') with Not_found -> None then
-    OpamConsole.warning "Command %S contains space characters" cmd;
   let full_cmd =
-    if resolve_path then t_resolve_command ~env ?dir cmd
+    if resolve_path then OpamStd.Sys.resolve_command ~env ?dir cmd
     else `Cmd cmd
   in
   match full_cmd with
@@ -520,17 +517,14 @@ let make_command
   | `Denied -> permission_denied cmd
 
 let run_process
-    ?verbose ?(env=default_env) ~name ?metadata ?stdout ?allow_stdin command =
+    ?verbose ?env ~name ?metadata ?stdout ?allow_stdin command =
+  let env = match env with None -> OpamProcess.default_env () | Some e -> e in
   let chrono = OpamConsole.timer () in
   runs := command :: !runs;
   match command with
   | []          -> invalid_arg "run_process"
   | cmd :: args ->
-
-    if OpamStd.String.contains_char cmd ' ' then
-      OpamConsole.warning "Command %S contains space characters" cmd;
-
-    match t_resolve_command ~env cmd with
+    match OpamStd.Sys.resolve_command ~env cmd with
     | `Cmd full_cmd ->
       let verbose = match verbose with
         | None   -> OpamCoreConfig.(!r.verbose_level) >= 2
@@ -576,11 +570,13 @@ let commands ?verbose ?env ?name ?metadata ?(keep_going=false) commands =
   | `Error e -> process_error e
   | `Exception e -> raise e
 
-let read_command_output ?verbose ?env ?metadata ?allow_stdin cmd =
+let read_command_output ?verbose ?env ?metadata ?allow_stdin
+    ?(ignore_stderr=false) cmd =
   let name = log_file None in
+  let stdout = name ^ (if ignore_stderr then ".stdout" else ".out") in
   let r =
     run_process ?verbose ?env ~name ?metadata ?allow_stdin
-      ~stdout:(name^".out")
+      ~stdout
       cmd
   in
   OpamProcess.cleanup r;
@@ -590,13 +586,7 @@ let read_command_output ?verbose ?env ?metadata ?allow_stdin cmd =
 let verbose_for_base_commands () =
   OpamCoreConfig.(!r.verbose_level) >= 3
 
-let cygify f =
-  if Sys.win32 then
-    List.map (Lazy.force f)
-  else
-    fun x -> x
-
-let copy_file src dst =
+let copy_file_t ?(with_log=true) src dst =
   if (try Sys.is_directory src
       with Sys_error _ -> raise (File_not_found src))
   then internal_error "Cannot copy %s: it is a directory." src;
@@ -605,30 +595,75 @@ let copy_file src dst =
   if file_or_symlink_exists dst
   then remove_file dst;
   mkdir (Filename.dirname dst);
-  log "copy %s -> %s" src dst;
+  if with_log || log_for_file_management () then
+    log "copy %s -> %s" src dst;
   copy_file_aux ~src ~dst ()
 
-let copy_dir src dst =
-  if Sys.file_exists dst then
-    if Sys.is_directory dst then
-      match ls src with
-      | [] -> ()
-      | srcfiles ->
-        command ~verbose:(verbose_for_base_commands ())
-          ([ "cp"; "-PRp" ] @ srcfiles @ [ dst ])
-    else internal_error "Can not copy dir %s to %s, which is not a directory"
-        src dst
-  else
-    (mkdir (Filename.dirname dst);
-     command ~verbose:(verbose_for_base_commands ())
-       [ "cp"; "-PRp"; src; dst ])
+let rec link_t ?(with_log=true) src dst =
+  mkdir (Filename.dirname dst);
+  if file_or_symlink_exists dst then
+    remove_file dst;
+  try
+    if with_log || log_for_file_management () then
+      log "ln -s %s %s" src dst;
+    Unix.symlink src dst
+  with Unix.Unix_error (Unix.EXDEV, _, _) ->
+    (* Fall back to copy if symlinks are not supported *)
+    let src =
+      if Filename.is_relative src then Filename.dirname dst / src
+      else src
+    in
+    if Sys.is_directory src then
+      copy_dir_t src dst
+    else
+      copy_file_t src dst
 
-let mv_aux f src dst =
+and copy_dir_t ?(with_log=true) src dst_dir =
+  if with_log || log_for_file_management () then
+    log "copydir %s -> %s" src dst_dir;
+  let files = get_files src in
+  mkdir dst_dir;
+  let with_log = false in
+  List.iter (fun file ->
+      let src = Filename.concat src file in
+      let dst = Filename.concat dst_dir file in
+      match Unix.lstat src with
+      | {Unix.st_kind = Unix.S_REG; _} ->
+        copy_file_t ~with_log src dst
+      | {Unix.st_kind = Unix.S_DIR; _} ->
+        copy_dir_t ~with_log src dst
+      | {Unix.st_kind = Unix.S_LNK; _} ->
+        let src = Unix.readlink src in
+        link_t ~with_log src dst
+      | {Unix.st_kind = Unix.S_CHR; _} ->
+        failwith (Printf.sprintf "Copying character devices (%s) is unsupported" src)
+      | {Unix.st_kind = Unix.S_BLK; _} ->
+        failwith (Printf.sprintf "Copying block devices (%s) is unsupported" src)
+      | {Unix.st_kind = Unix.S_FIFO; _} ->
+        failwith (Printf.sprintf "Copying named pipes (%s) is unsupported" src)
+      | {Unix.st_kind = Unix.S_SOCK; _} ->
+        failwith (Printf.sprintf "Copying sockets (%s) is unsupported" src)
+      | exception Unix.(Unix_error (ENOENT, _, _)) when Sys.win32 ->
+        (* This is usually caused by invalid symlinks extracted by a
+           Cygwin/MSYS2 tar. *)
+        OpamConsole.warning "Warning: cannot copy %s to %s" src dst_dir
+    ) files
+
+let copy_dir = copy_dir_t ~with_log:true
+let copy_file = copy_file_t ~with_log:true
+
+let mv src dst =
   if file_or_symlink_exists dst then remove_file dst;
   mkdir (Filename.dirname dst);
-  command ~verbose:(verbose_for_base_commands ()) ("mv"::(cygify f [src; dst]))
-
-let mv = mv_aux (get_cygpath_function ~command:"mv")
+  log "mv %s -> %s" src dst;
+  try
+    Unix.rename src dst
+  with
+  | Unix.Unix_error(Unix.EXDEV, _, _) ->
+    let with_log = false in
+    if Sys.is_directory src
+    then (copy_dir_t ~with_log src dst; remove_dir_t src)
+    else (copy_file_t ~with_log src dst; remove_file_t ~with_log src)
 
 let is_exec file =
   let stat = Unix.stat file in
@@ -704,21 +739,33 @@ let classify_executable file =
     `Unknown
 
 let default_install_warning dst = function
-| `Add_exe ->
+  | `Add_exe ->
     OpamConsole.warning "Automatically adding .exe to %s" dst
-| `Install_dll ->
-    (* TODO Installation of .dll to bin is unfortunate, but not sure if it should be a warning *)
+  | `Install_dll ->
+    (* TODO Installation of .dll to bin is unfortunate, but not sure if it
+       should be a warning *)
     ()
-| `Install_script ->
+  | `Install_script ->
     (* TODO Generate a .cmd wrapper (and warn about it - they're not perfect) *)
     OpamConsole.warning "%s is a script; the command won't be available" dst;
-| `Install_unknown ->
-    (* TODO Installation of a non-executable file is unexpected, but not sure if it should be a warning/error *)
+  | `Install_unknown ->
+    (* TODO Installation of a non-executable file is unexpected, but not sure
+       if it should be a warning/error *)
     ()
-| `Cygwin ->
+  | `Cygwin ->
     OpamConsole.warning "%s is a Cygwin-linked executable" dst
-| `Cygwin_libraries ->
-    OpamConsole.warning "%s links with a Cygwin-compiled DLL (almost certainly a packaging or environment error)" dst
+  | `Msys2 ->
+    OpamConsole.warning "%s is a MSYS2-linked executable" dst
+  | `Tainted `Cygwin ->
+    OpamConsole.warning
+      "%s is an executable which links to a Cygwin-linked library" dst
+  | `Tainted `Msys2 ->
+    OpamConsole.warning
+      "%s is an executable which links to a MSYS2-linked library" dst
+  | `Cygwin_libraries ->
+    OpamConsole.warning
+      "%s links with a Cygwin-compiled DLL (almost certainly a packaging \
+       or environment error)" dst
 
 let install ?(warning=default_install_warning) ?exec src dst =
   if Sys.is_directory src then
@@ -753,13 +800,10 @@ let install ?(warning=default_install_warning) ?exec src dst =
       in
       copy_file_aux ~src ~dst ();
       if cygcheck then
-        match OpamStd.Sys.is_cygwin_variant dst with
-          `Native ->
-            ()
-        | `Cygwin ->
-            warning dst `Cygwin
-        | `CygLinked ->
-            warning dst `Cygwin_libraries
+        match OpamStd.Sys.get_windows_executable_variant
+                ?search_in_first:OpamCoreConfig.(!r.cygbin) dst with
+        | `Native -> ()
+        | (`Cygwin | `Msys2 | `Tainted _) as code -> warning dst code
     end else
       copy_file_aux ~src ~dst ()
   else
@@ -836,12 +880,8 @@ module Tar = struct
              else None)
         None
         extensions in
-    match ext with
-    | Some t -> Some t
-    | None   ->
-      match guess_type file with
-      | Some t -> Some t
-      | _ -> None
+    if Sys.file_exists file then guess_type file
+    else ext
 
   let is_archive file =
     get_type file <> None
@@ -889,18 +929,22 @@ module Tar = struct
 end
 
 module Zip = struct
+
   let is_archive f =
-    try
-      let ic = open_in f in
-      let c1 = input_char ic in
-      let c2 = input_char ic in
-      let c3 = input_char ic in
-      let c4 = input_char ic in
-      close_in ic;
-      match c1, c2, c3, c4 with
-      | '\x50', '\x4b', '\x03', '\x04' -> true
-      | _ -> false
-    with Sys_error _ | End_of_file -> false
+    if Sys.file_exists f then
+      try
+        let ic = open_in f in
+        let c1 = input_char ic in
+        let c2 = input_char ic in
+        let c3 = input_char ic in
+        let c4 = input_char ic in
+        close_in ic;
+        match c1, c2, c3, c4 with
+        | '\x50', '\x4b', '\x03', '\x04' -> true
+        | _ -> false
+      with Sys_error _ | End_of_file -> false
+    else
+      Filename.check_suffix f "zip"
 
   let extract_command file =
     Some (fun dir -> make_command "unzip" [ file; "-d"; dir ])
@@ -994,13 +1038,7 @@ let extract_in ~dir file =
   | None -> ()
 
 let link src dst =
-  mkdir (Filename.dirname dst);
-  if file_or_symlink_exists dst then
-    remove_file dst;
-  try
-    log "ln -s %s %s" src dst;
-    Unix.symlink src dst
-  with Unix.Unix_error (Unix.EXDEV, _, _) ->
+  let fallback () =
     (* Fall back to copy if symlinks are not supported *)
     let src =
       if Filename.is_relative src then Filename.dirname dst / src
@@ -1010,6 +1048,23 @@ let link src dst =
       copy_dir src dst
     else
       copy_file src dst
+  in
+  mkdir (Filename.dirname dst);
+  if file_or_symlink_exists dst then (
+    (* TODO: move this into remove_file *)
+    if Sys.win32 then Unix.chmod dst 0o640;
+    remove_file dst
+  );
+  if Unix.has_symlink () then
+    try
+      log "ln -s %s %s" src dst;
+      Unix.symlink src dst
+    with Unix.Unix_error (Unix.EXDEV, _, _) ->
+      fallback ()
+  else (
+    log "copy %s -> %s" src dst;
+    fallback ()
+  )
 
 type actual_lock_flag = [ `Lock_read | `Lock_write ]
 type lock_flag = [ `Lock_none | actual_lock_flag ]
@@ -1072,20 +1127,21 @@ let rec flock_update
          if Sys.win32 && kind <> `Lock_none then
            Unix.(lockf fd F_ULOCK 0);
          Unix.lockf fd (unix_lock_op ~dontblock:true flag) 0
-       with Unix.Unix_error (Unix.EAGAIN,_,_) ->
+       with Unix.Unix_error (Unix.EAGAIN,_,_)
+          | Unix.Unix_error (Unix.EACCES,_,_) ->
          if dontblock then
            OpamConsole.error_and_exit `Locked
              "Another process has locked %s and non blocking mode enabled"
              file;
-         OpamConsole.formatted_msg
+         OpamConsole.formatted_errmsg
            "Another process has locked %s, waiting (%s to abort)... "
            file (if Sys.win32 then "CTRL+C" else "C-c");
          let rec lock_w_ignore_sig () =
            try Unix.lockf fd (unix_lock_op ~dontblock:false flag) 0;
-           with Sys.Break as e -> (OpamConsole.msg "\n"; raise e)
+           with Sys.Break as e -> (OpamConsole.errmsg "\n"; raise e)
               | Unix.Unix_error (Unix.EINTR,_,_) -> lock_w_ignore_sig ()
          in lock_w_ignore_sig ();
-         OpamConsole.msg "lock acquired.\n");
+         OpamConsole.errmsg "lock acquired.\n");
     lock.kind <- (flag :> lock_flag)
   | _ -> assert false
 
@@ -1098,7 +1154,7 @@ and flock: 'a. ([< lock_flag ] as 'a) -> ?dontblock:bool -> string -> lock =
   | flag ->
     mkdir (Filename.dirname file);
     let rdflag = if (flag :> lock_flag) = `Lock_write then Unix.O_RDWR else Unix.O_RDONLY in
-    let fd = Unix.openfile file Unix.([O_CREAT; O_CLOEXEC; rdflag]) 0o666 in
+    let fd = Unix.openfile file Unix.([O_CREAT; O_CLOEXEC; O_SHARE_DELETE; rdflag]) 0o666 in
     Hashtbl.add locks fd ();
     let lock = { fd = Some fd; file; kind = `Lock_none } in
     flock_update flag ?dontblock lock;
@@ -1488,6 +1544,40 @@ let translate_patch ~dir orig corrected =
   end;
   close_in ch
 
+let gpatch = lazy begin
+  let rec search_gpatch = function
+    | [] -> None
+    | patch_cmd::patch_cmds ->
+      match OpamProcess.run (make_command ~name:"patch" patch_cmd ["--version"]) with
+      | r ->
+        (match OpamProcess.is_success r, r.OpamProcess.r_stdout with
+         | true, full::_ when
+             OpamStd.String.is_prefix_of ~from:0 ~full "GNU patch " ->
+           Some patch_cmd
+         | _ ->
+           search_gpatch patch_cmds)
+      | exception _ -> search_gpatch patch_cmds
+  in
+  let default_cmd, other_cmds =
+    match OpamStd.Sys.os () with
+    | DragonFly
+    | FreeBSD
+    | NetBSD
+    | OpenBSD -> ("gpatch", ["patch"])
+    | Cygwin
+    | Darwin
+    | Linux
+    | Unix
+    | Win32
+    | Other _ -> ("patch", ["gpatch"])
+  in
+  match search_gpatch (default_cmd :: other_cmds) with
+  | Some gpatch -> gpatch
+  | None ->
+    OpamConsole.warning "Invalid patch utility. Please install GNU patch";
+    default_cmd
+end
+
 let patch ?(preprocess=true) ~dir p =
   if not (Sys.file_exists p) then
     (OpamConsole.error "Patch file %S not found." p;
@@ -1500,12 +1590,7 @@ let patch ?(preprocess=true) ~dir p =
     else
       p
   in
-  let patch_cmd =
-    match OpamStd.Sys.os () with
-    | OpamStd.Sys.OpenBSD
-    | OpamStd.Sys.FreeBSD -> "gpatch"
-    | _ -> "patch"
-  in
+  let patch_cmd = Lazy.force gpatch in
   make_command ~name:"patch" ~dir patch_cmd ["-p1"; "-i"; p'] @@> fun r ->
     if not (OpamConsole.debug ()) then Sys.remove p';
     if OpamProcess.is_success r then Done None
@@ -1531,6 +1616,3 @@ let init () =
   Sys.catch_break true;
   try Sys.set_signal Sys.sigpipe (Sys.Signal_handle (fun _ -> ()))
   with Invalid_argument _ -> ()
-
-let () =
-  OpamProcess.set_resolve_command resolve_command
