@@ -1,6 +1,6 @@
 (**************************************************************************)
 (*                                                                        *)
-(*    Copyright 2012-2015 OCamlPro                                        *)
+(*    Copyright 2012-2020 OCamlPro                                        *)
 (*    Copyright 2012 INRIA                                                *)
 (*                                                                        *)
 (*  All rights reserved. This file is distributed under the terms of the  *)
@@ -12,6 +12,7 @@
 (** Handles all OPAM file formats as record types and submodules, conversion to
     and from syntax *)
 
+open OpamParserTypes.FullPos
 open OpamTypes
 
 (** Functions to read and write OPAM configuration files in a typed way *)
@@ -32,6 +33,8 @@ module type IO_FILE = sig
   (** File contents *)
   type t
 
+  val format_version: OpamVersion.t
+
   (** Empty file *)
   val empty: t
 
@@ -57,6 +60,16 @@ module type IO_FILE = sig
 
 end
 
+(* Error less [IO_FILE] read functions. *)
+module type BestEffortRead = sig
+  type t
+  val read: t typed_file -> t
+  val read_opt: t typed_file -> t option
+  val safe_read: t typed_file -> t
+  val read_from_channel: ?filename:t typed_file -> in_channel -> t
+  val read_from_string: ?filename:t typed_file -> string -> t
+end
+
 (** Lines of space-separated words. *)
 module Lines: IO_FILE with type t = string list list
 
@@ -88,6 +101,18 @@ module Wrappers: sig
   val pre_session: t -> command list
   val post_session: t -> command list
 
+  val with_pre_build: command list -> t -> t
+  val with_wrap_build: command list -> t -> t
+  val with_post_build: command list -> t -> t
+  val with_pre_install: command list -> t -> t
+  val with_wrap_install: command list -> t -> t
+  val with_post_install: command list -> t -> t
+  val with_pre_remove: command list -> t -> t
+  val with_wrap_remove: command list -> t -> t
+  val with_post_remove: command list -> t -> t
+  val with_pre_session: command list -> t -> t
+  val with_post_session: command list -> t -> t
+
   val empty : t
 
   val add: outer:t -> inner:t -> t
@@ -98,10 +123,14 @@ module Config: sig
 
   include IO_FILE
 
+  (** Current root version *)
+  val root_version: opam_version
+
   (** OCaml switch updates *)
   val with_switch: switch -> t -> t
   val with_switch_opt: switch option -> t -> t
 
+  (** The list of switches, ordered by most recently used first *)
   val with_installed_switches: switch list -> t -> t
 
   (** Repository updates *)
@@ -109,14 +138,17 @@ module Config: sig
 
   (** Update opam-version *)
   val with_opam_version: OpamVersion.t -> t -> t
+  val with_opam_root_version: OpamVersion.t -> t -> t
 
   val with_criteria: (solver_criteria * string) list -> t -> t
   val with_best_effort_prefix: string -> t -> t
+  val with_best_effort_prefix_opt: string option-> t -> t
 
   val with_solver: arg list -> t -> t
   val with_solver_opt: arg list option -> t -> t
 
   val with_jobs: int -> t -> t
+  val with_jobs_opt: int option -> t -> t
   val with_dl_tool: arg list -> t -> t
   val with_dl_tool_opt: arg list option -> t -> t
   val with_dl_jobs: int -> t -> t
@@ -131,20 +163,40 @@ module Config: sig
     arg list option -> t -> t
   val with_default_compiler:
     formula -> t -> t
+  val with_default_invariant:
+    formula -> t -> t
+  val with_depext: bool -> t -> t
+  val with_depext_run_installs: bool -> t -> t
+  val with_depext_cannot_install: bool -> t -> t
+  val with_depext_bypass: OpamSysPkg.Set.t -> t -> t
+  val with_sys_pkg_manager_cmd: filename OpamStd.String.Map.t -> t -> t
 
-  (** Return the OPAM version *)
+  val with_swh_fallback: bool -> t -> t
+  val with_git_location: dirname -> t -> t
+  val with_git_location_opt: dirname option -> t -> t
+
+  (** Return the opam version *)
   val opam_version: t  -> opam_version
 
+  (** Return the opam root version, if not defined returns the default version
+      value [2.1~~previous] *)
+  val opam_root_version: t -> opam_version
+
+  (** Return the opam root version if defined *)
+  val opam_root_version_opt: t -> opam_version option
+
   (** Return the list of repository *)
   val repositories: t  -> repository_name list
 
   (** Return the OCaml switch *)
   val switch: t -> switch option
 
+  val previous_switch : t -> switch option
+
   val installed_switches: t -> switch list
 
-  (** Return the number of jobs *)
-  val jobs: t -> int
+  (** Return the number of jobs defined *)
+  val jobs: t -> int option
 
   val dl_tool: t -> arg list option
 
@@ -170,7 +222,31 @@ module Config: sig
   val validation_hook: t -> arg list option
 
   val default_compiler: t -> formula
+  val default_invariant: t -> formula
 
+  val depext: t -> bool
+  val depext_run_installs: t -> bool
+  val depext_cannot_install: t -> bool
+  val depext_bypass: t -> OpamSysPkg.Set.t
+
+  val sys_pkg_manager_cmd: t -> filename OpamStd.String.Map.t
+
+  (* Enable or disable Software Heritage fallback for unreachable package
+     sources *)
+  val swh_fallback: t -> bool
+
+  val git_location: t -> dirname option
+
+  val fields: (string * (t, value) OpamPp.field_parser) list
+
+  (** All file fields as print-AST, Fields within sections are
+      accessed through dot-separated paths *)
+  val to_list: ?filename:'a typed_file -> t -> (string * value) list
+
+  module BestEffort: BestEffortRead with type t := t
+
+  (** Raw read the config file to extract [opam-root-version] field value. *)
+  val raw_root_version: 'a typed_file -> OpamVersion.t option
 end
 
 (** Init config file [/etc/opamrc] *)
@@ -180,6 +256,7 @@ module InitConfig: sig
   val opam_version: t -> opam_version
   val repositories: t -> (repository_name * (url * trust_anchors option)) list
   val default_compiler: t -> formula
+  val default_invariant: t -> formula
   val jobs: t -> int option
   val dl_tool: t -> arg list option
   val dl_jobs: t -> int option
@@ -192,11 +269,14 @@ module InitConfig: sig
   val recommended_tools: t -> (string list * string option * filter option) list
   val required_tools: t -> (string list * string option * filter option) list
   val init_scripts: t -> ((string * string) * filter option) list
+  val sys_pkg_manager_cmd: t -> filename OpamStd.String.Map.t
+  val git_location: t -> dirname option
 
   val with_opam_version: opam_version -> t -> t
   val with_repositories:
     (repository_name * (url * trust_anchors option)) list -> t -> t
   val with_default_compiler: formula -> t -> t
+  val with_default_invariant: formula -> t -> t
   val with_jobs: int option -> t -> t
   val with_dl_tool: arg list option -> t -> t
   val with_dl_jobs: int option -> t -> t
@@ -241,7 +321,10 @@ module URL: sig
 
   include IO_FILE
 
-  val create: ?mirrors:url list -> ?checksum:OpamHash.t list -> url -> t
+  val create:
+    ?mirrors:url list -> ?checksum:OpamHash.t list ->
+    ?swhid:OpamSWHID.t -> ?subpath:subpath ->
+    url -> t
 
   (** URL address *)
   val url: t -> url
@@ -250,9 +333,18 @@ module URL: sig
 
   (** Archive checksum *)
   val checksum: t -> OpamHash.t list
+  val swhid: t -> OpamSWHID.t option
 
   (** Constructor *)
+  val with_url: url -> t -> t
   val with_checksum: OpamHash.t list -> t -> t
+  val with_mirrors: OpamUrl.t list -> t -> t
+  val with_swhid: OpamSWHID.t -> t -> t
+  val with_swhid_opt: OpamSWHID.t option -> t -> t
+  val with_subpath: subpath -> t -> t
+  val with_subpath_opt: subpath option -> t -> t
+
+  val subpath: t -> subpath option
 
 end
 
@@ -273,7 +365,7 @@ module OPAM: sig
     conflict_class : name list;
     available  : filter;
     flags      : package_flag list;
-    env        : env_update list;
+    env        : (spf_unresolved, euok_writeable) env_update list;
 
     (* Build instructions *)
     build      : command list;
@@ -284,14 +376,14 @@ module OPAM: sig
     (* Auxiliary data affecting the build *)
     substs     : basename list;
     patches    : (basename * filter option) list;
-    build_env  : env_update list;
+    build_env  : (spf_unresolved, euok_writeable) env_update list;
     features   : (OpamVariable.t * filtered_formula * string) list;
     extra_sources: (basename * URL.t) list;
 
     (* User-facing data used by opam *)
     messages   : (string * filter option) list;
     post_messages: (string * filter option) list;
-    depexts    : (string list * filter) list;
+    depexts    : (OpamSysPkg.Set.t * filter) list;
     libraries  : (string * filter option) list;
     syntax     : (string * filter option) list;
     dev_repo   : url option;
@@ -307,19 +399,24 @@ module OPAM: sig
     bug_reports: string list;
 
     (* Extension fields (x-foo: "bar") *)
-    extensions : (pos * value) OpamStd.String.Map.t;
+    extensions : value OpamStd.String.Map.t;
 
     (* Extra sections *)
     url        : URL.t option;
     descr      : Descr.t option;
 
     (* Related metadata directory (not an actual field of the file)
-       This can be used to locate e.g. the files/ overlays *)
-    metadata_dir: dirname option;
+       This can be used to locate e.g. the files/ overlays.
+       If the repository is specified, the string is a relative path from its
+       root. It should otherwise be an absolute path. *)
+    metadata_dir: (repository_name option * string) option;
 
     (* Names and hashes of the files below files/ *)
     extra_files: (OpamFilename.Base.t * OpamHash.t) list option;
 
+    (* Origin with the extension *)
+    locked: string option;
+
     format_errors: (string * OpamPp.bad_format) list;
 
     (* Deprecated, for compat and proper linting *)
@@ -338,12 +435,16 @@ module OPAM: sig
 
   (** Returns the opam value (including url, descr) with all non-effective (i.e.
       user-directed information that doesn't change opam's view on the package)
-      fields set to their empty values. Useful for comparisons. *)
-  val effective_part: t -> t
+      fields set to their empty values. Useful for comparisons.
+      @param ?modulo_state if [true], eliminates the fields relying on the state
+      of the switch (depends, available, …). This is [false] by default. *)
+  val effective_part: ?modulo_state:bool -> t -> t
 
   (** Returns true if the effective parts of the two package definitions are
-      equal *)
-  val effectively_equal: t -> t -> bool
+      equal.
+      @param ?modulo_state if [true], considers the fields relying on the state
+      of the switch (depends, available, …) equal. This is [false] by default *)
+  val effectively_equal: ?modulo_state:bool -> t -> t -> bool
 
   (** Compares two package definitions, ignoring the virtual fields bound to
       file location ([metadata_dir]...) *)
@@ -376,7 +477,7 @@ module OPAM: sig
   val substs: t -> basename list
 
   (** List of environment variables to set-up for the build *)
-  val build_env: t -> env_update list
+  val build_env: t -> (spf_unresolved, [> euok_writeable]) env_update list
 
   (** List of command to run for building the package *)
   val build: t -> command list
@@ -394,7 +495,7 @@ module OPAM: sig
   val depopts: t -> filtered_formula
 
   (** External dependencies *)
-  val depexts: t -> (string list * filter) list
+  val depexts: t -> (OpamSysPkg.Set.t * filter) list
 
   val extra_sources: t -> (basename * URL.t) list
 
@@ -466,7 +567,7 @@ module OPAM: sig
   val has_flag: package_flag -> t -> bool
 
   (** The environment variables that this package exports *)
-  val env: t -> env_update list
+  val env: t -> (spf_unresolved, [> euok_writeable]) env_update list
 
   val descr: t -> Descr.t option
 
@@ -477,17 +578,28 @@ module OPAM: sig
 
   val get_url: t -> url option
 
-  (** Related metadata directory (not an actual field of the file, linked to the
-      file location).
+  (** Related metadata directory (either repository name + relative path, or
+      absolute path; not an actual field of the file, linked to the file
+      location).
       This can be used to locate e.g. the files/ overlays *)
-  val metadata_dir: t -> dirname option
+  val metadata_dir: t -> (repository_name option * string) option
+
+  (** Gets the resolved metadata dir, given a mapping of repository names to
+      their roots *)
+  val get_metadata_dir:
+    repos_roots:(repository_name -> dirname) -> t -> dirname option
 
   (** Names and hashes of the files below files/ *)
   val extra_files: t -> (OpamFilename.Base.t * OpamHash.t) list option
 
+  (** From locked opam file *)
+  val locked: t -> string option
+
   (** Looks up the extra files, and returns their full paths, relative path to
       the package source, and hash. Doesn't check the hashes. *)
-  val get_extra_files: t -> (filename * basename * OpamHash.t) list
+  val get_extra_files:
+    repos_roots:(repository_name -> dirname) ->
+    t -> (filename * basename * OpamHash.t) list
 
   (** Returns the errors that were found when parsing the file, associated to
       their fields (that were consequently ignored) *)
@@ -540,7 +652,7 @@ module OPAM: sig
   (** Construct as [substs] *)
   val with_substs: basename list -> t -> t
 
-  val with_build_env: env_update list -> t -> t
+  val with_build_env: (spf_unresolved, euok_writeable) env_update list -> t -> t
 
   val with_available: filter -> t -> t
 
@@ -560,7 +672,7 @@ module OPAM: sig
   val with_bug_reports: string list -> t -> t
 
   (** Construct using [depexts] *)
-  val with_depexts: (string list * filter) list -> t -> t
+  val with_depexts: (OpamSysPkg.Set.t * filter) list -> t -> t
 
   val with_flags: package_flag list -> t -> t
 
@@ -568,7 +680,7 @@ module OPAM: sig
 
   val with_tags: string list -> t -> t
 
-  val with_env: env_update list -> t -> t
+  val with_env: (spf_unresolved, euok_writeable) env_update list -> t -> t
 
   val with_dev_repo: url -> t -> t
 
@@ -582,6 +694,8 @@ module OPAM: sig
 
   val add_extension: t -> string -> value -> t
 
+  val remove_extension: t -> string -> t
+
   val with_deprecated_build_doc: command list -> t -> t
 
   val with_deprecated_build_test: command list -> t -> t
@@ -597,11 +711,13 @@ module OPAM: sig
   val with_url: URL.t -> t -> t
   val with_url_opt: URL.t option -> t -> t
 
-  val with_metadata_dir: dirname option -> t -> t
+  val with_metadata_dir: (repository_name option * string) option -> t -> t
 
   val with_extra_files: (OpamFilename.Base.t * OpamHash.t) list -> t -> t
   val with_extra_files_opt: (OpamFilename.Base.t * OpamHash.t) list option -> t -> t
 
+  val with_locked_opt: string option -> t -> t
+
   val with_format_errors: (string * OpamPp.bad_format) list -> t -> t
 
   (** Prints to a string, while keeping the format of the original file as much
@@ -644,6 +760,9 @@ module OPAM: sig
       sections can be accessed through [section.field]. *)
   val print_field_as_syntax: string -> t -> value option
 
+  (** x-field name for path rewriting on windows *)
+  val rewrite_xfield: string
+
 end
 
 (** Compiler aliases: [$opam/aliases]. Deprecated, used only for migration *)
@@ -662,6 +781,7 @@ end
 module SwitchSelections: sig
   type t = switch_selections
   include IO_FILE with type t := t
+  module BestEffort: BestEffortRead with type t := t
 end
 
 (** An extended version of SwitchSelections that can include full opam files as
@@ -669,6 +789,7 @@ end
 module SwitchExport: sig
   type t = {
     selections: switch_selections;
+    extra_files: string OpamHash.Map.t;
     overlays: OPAM.t OpamPackage.Name.Map.t;
   }
   include IO_FILE with type t := t
@@ -680,7 +801,15 @@ end
 module PkgList: IO_FILE with type t = package_set
 
 (** Cached environment updates (<switch>/environment) *)
-module Environment: IO_FILE with type t = env_update list
+module Environment : sig
+  include IO_FILE with type t = (spf_resolved, euok_writeable) env_update list
+
+  val read: t typed_file -> (spf_resolved, [> euok_writeable ]) env_update list
+  val read_opt: t typed_file -> (spf_resolved, [> euok_writeable ]) env_update list option
+  val safe_read: t typed_file -> (spf_resolved, [> euok_writeable ]) env_update list
+  val read_from_channel: ?filename:t typed_file -> in_channel -> (spf_resolved, [> euok_writeable ]) env_update list
+  val read_from_string: ?filename:t typed_file -> string -> (spf_resolved, [> euok_writeable ]) env_update list
+end
 
 (** Compiler version [$opam/compilers/]. Deprecated, only used to upgrade old
     data *)
@@ -693,7 +822,7 @@ module Comp: sig
 
   (** Create a pre-installed compiler description file *)
   val create_preinstalled:
-    compiler -> compiler_version -> name list -> env_update list -> t
+    compiler -> compiler_version -> name list -> (spf_unresolved, euok_writeable) env_update list -> t
 
   (** Is it a pre-installed compiler description file *)
   val preinstalled: t -> bool
@@ -728,7 +857,7 @@ module Comp: sig
 
   (** Environment variable to set-up before running commands in the
       subtree *)
-  val env: t -> env_update list
+  val env: t -> (spf_unresolved, euok_writeable) env_update list
 
   val tags: t -> string list
 
@@ -799,6 +928,48 @@ module Dot_install: sig
   (** List of other files to install *)
   val misc: t -> (basename optional * filename) list
 
+  (** List of files to install in $bin/ *)
+  val with_bin : (basename optional * basename option) list -> t -> t
+
+  (** List of files to install in $sbin/ *)
+  val with_sbin : (basename optional * basename option) list -> t -> t
+
+  (** List of files to install in $lib/ *)
+  val with_lib : (basename optional * basename option) list -> t -> t
+
+  (** List of toplevel files *)
+  val with_toplevel : (basename optional * basename option) list -> t -> t
+
+  (** C bindings *)
+  val with_stublibs : (basename optional * basename option) list -> t -> t
+
+  (** List of architecture-independent files *)
+  val with_share : (basename optional * basename option) list -> t -> t
+
+  (** List of files under the more general share prefix *)
+  val with_share_root : (basename optional * basename option) list -> t -> t
+
+  (** List of etc files *)
+  val with_etc : (basename optional * basename option) list -> t -> t
+
+  (** List of doc files *)
+  val with_doc : (basename optional * basename option) list -> t -> t
+
+  (** Man pages *)
+  val with_man : (basename optional * basename option) list -> t -> t
+
+  (** Executable files under lib/ *)
+  val with_libexec : (basename optional * basename option) list -> t -> t
+
+  (** Not relative to the package's lib dir *)
+  val with_lib_root : (basename optional * basename option) list -> t -> t
+
+  (** Not relative to the package's lib dir, and with +x set *)
+  val with_libexec_root : (basename optional * basename option) list -> t -> t
+
+  (** List of other files to install *)
+  val with_misc : (basename optional * filename) list -> t -> t
+
 end
 
 (** .changes files, bound to the OpamDirTrack module *)
@@ -856,9 +1027,11 @@ module Repo_config_legacy : sig
   include IO_FILE with type t := t
 end
 
-
-module Repos_config: IO_FILE
-  with type t = (url * trust_anchors option) option OpamRepositoryName.Map.t
+module Repos_config: sig
+  type t = (url * trust_anchors option) option OpamRepositoryName.Map.t
+  include IO_FILE with type t := t
+  module BestEffort: BestEffortRead with type t := t
+end
 
 module Switch_config: sig
   type t = {
@@ -869,12 +1042,24 @@ module Switch_config: sig
     variables: (variable * variable_contents) list;
     opam_root: dirname option;
     wrappers: Wrappers.t;
-    env: env_update list;
+    env: (spf_resolved, euok_writeable) env_update list;
+    invariant: OpamFormula.t option;
+    depext_bypass: OpamSysPkg.Set.t;
   }
+
+  val env: t -> (spf_resolved, [> euok_writeable ]) env_update list
+
+  val file_format_version: OpamVersion.t
   val variable: t -> variable -> variable_contents option
   val path: t -> std_path -> string option
   val wrappers: t -> Wrappers.t
+  val sections: (string * (t, (string option * opamfile_item list) list) OpamPp.field_parser) list
+  val fields: (string * (t, value) OpamPp.field_parser) list
+  val to_list: ?filename:'a typed_file -> t -> (string * value) list
   include IO_FILE with type t := t
+  val oldest_compatible_format_version: OpamVersion.t
+
+  module BestEffort: BestEffortRead with type t := t
 end
 
 (** Pinned package files (only used for migration from 1.2, the inclusive State
@@ -977,6 +1162,8 @@ end
 
 module type SyntaxFileArg = sig
   val internal: string
+  val atomic: bool
+  val format_version: OpamVersion.t
   type t
   val empty: t
   val pp: (opamfile, filename * t) OpamPp.t
@@ -986,6 +1173,7 @@ module SyntaxFile(X: SyntaxFileArg) : IO_FILE with type t := X.t
 
 module type LineFileArg = sig
   val internal: string
+  val atomic: bool
   type t
   val empty: t
   val pp: (string list list, t) OpamPp.t
