@@ -50,6 +50,7 @@ module VCS : OpamVCS.VCS = struct
        *)
       git repo_root [ "config" ; "--local" ; "core.autocrlf"; "false"];
       git repo_root [ "config" ; "--local" ; "core.eol"; "lf"];
+      git repo_root [ "config" ; "--local" ; "color.ui"; "false" ];
       (* Document the remote for user-friendliness (we don't use it) *)
       git repo_root [ "remote"; "add"; "origin"; OpamUrl.base_url repo_url ];
     ] @@+ function
@@ -61,12 +62,13 @@ module VCS : OpamVCS.VCS = struct
     | Some h -> "refs/remotes/opam-ref-"^h
     | None -> "refs/remotes/opam-ref"
 
-  let fetch ?cache_dir ?subpath repo_root repo_url =
+  let fetch ?(full_fetch = false) ?cache_dir ?subpath repo_root repo_url =
     (match subpath with
      | Some sp ->
        git repo_root [ "config"; "--local"; "core.sparseCheckout"; "true" ]
        @@> fun r -> OpamSystem.raise_on_process_error r;
-       OpamFilename.write (repo_root / ".git" / "info" // "sparse-checkout") sp;
+       OpamFilename.write (repo_root / ".git" / "info" // "sparse-checkout")
+         (OpamFilename.SubPath.normalised_string sp);
        Done()
      | None -> Done())
     @@+ fun _ ->
@@ -93,7 +95,10 @@ module VCS : OpamVCS.VCS = struct
           OpamFilename.write alternates
             (OpamFilename.Dir.to_string (cache / "objects")))
       global_cache;
-    git repo_root [ "fetch" ; "-q"; origin; "--update-shallow"; refspec ]
+    (if full_fetch then
+       git repo_root [ "fetch" ; "-q"; origin; "--update-shallow"; refspec ]
+     else
+       git repo_root [ "fetch" ; "-q"; origin; "--update-shallow"; "--depth=1"; refspec ])
     @@> fun r ->
     if OpamProcess.check_success_and_cleanup r then
       let refspec =
@@ -125,11 +130,11 @@ module VCS : OpamVCS.VCS = struct
         (* the above might still fail on raw, untracked hashes: try to bind to
            the direct refspec, if found *)
         (git repo_root [ "update-ref" ; opam_ref; branch ] @@> fun r ->
-          if OpamProcess.check_success_and_cleanup r then
-            Done()
-          else
-            (* check if the commit exists *)
-            (git repo_root [ "fetch"; "-q" ] @@> fun r ->
+         if OpamProcess.check_success_and_cleanup r then
+           Done()
+         else
+           (* check if the commit exists *)
+           (git repo_root [ "fetch"; "-q" ] @@> fun r ->
             OpamSystem.raise_on_process_error r;
             git repo_root [ "show"; "-s"; "--format=%H"; branch ] @@> fun r ->
             if OpamProcess.check_success_and_cleanup r then
@@ -137,14 +142,12 @@ module VCS : OpamVCS.VCS = struct
             else failwith "Commit not found on repository"))
       else
       let error = r in
-      git repo_root ["ls-files"] @@> function
+      git repo_root ["branch"; "-a"] @@> function
       | { OpamProcess.r_code = 0; OpamProcess.r_stdout = []; _ } ->
-        git repo_root ["show"] @@> fun r ->
-        if OpamProcess.is_failure r then
-          failwith "Git repository seems just initialized, \
-                    try again after your first commit"
-        else
-          OpamSystem.process_error error
+        failwith "Git repository seems just initialized, \
+                  try again after your first commit"
+      | { OpamProcess.r_code = 0; OpamProcess.r_stdout = _::_; _ } ->
+        failwith (Printf.sprintf "Branch %s not found" branch)
       | _ -> OpamSystem.process_error error
 
   let revision repo_root =
@@ -157,10 +160,7 @@ module VCS : OpamVCS.VCS = struct
        match r.OpamProcess.r_stdout with
        | []      -> Done None
        | full::_ ->
-         if String.length full > 8 then
-           Done (Some (String.sub full 0 8))
-         else
-           Done (Some full))
+         Done (Some full))
 
   let clean repo_root =
     git repo_root [ "clean"; "-fdx" ]
@@ -210,9 +210,11 @@ module VCS : OpamVCS.VCS = struct
     else
       Done (Some (OpamFilename.of_string patch_file))
 
-  let is_up_to_date repo_root repo_url =
+  let is_up_to_date ?subpath repo_root repo_url =
     let rref = remote_ref repo_url in
-    git repo_root [ "diff" ; "--no-ext-diff" ; "--quiet" ; rref; "--" ]
+    git repo_root ([ "diff" ; "--no-ext-diff" ; "--quiet" ; rref; "--" ]
+                   @ List.map OpamFilename.SubPath.to_string
+                     (OpamStd.Option.to_list subpath))
     @@> function
     | { OpamProcess.r_code = 0; _ } -> Done true
     | { OpamProcess.r_code = 1; _ } as r ->
@@ -222,7 +224,7 @@ module VCS : OpamVCS.VCS = struct
   let versioned_files repo_root =
     git repo_root ~verbose:false [ "ls-files" ] @@> fun r ->
     OpamSystem.raise_on_process_error r;
-    Done r.OpamProcess.r_stdout
+    Done (List.map OpamSystem.forward_to_back r.OpamProcess.r_stdout)
 
   let vc_dir repo_root = OpamFilename.Op.(repo_root / ".git")
 
@@ -238,7 +240,8 @@ module VCS : OpamVCS.VCS = struct
     let subpath =
       match subpath with
       | None -> []
-      | Some dir -> ["--" ; dir] in
+      | Some dir -> ["--" ; OpamFilename.SubPath.to_string dir]
+      in
     git dir ([ "diff"; "--no-ext-diff"; "--quiet" ; "HEAD" ] @ subpath)
     @@> function
     | { OpamProcess.r_code = 0; _ } ->
@@ -271,13 +274,6 @@ module VCS : OpamVCS.VCS = struct
 
   (** check if a hash or branch is present in remote origin and returns  *)
   let check_remote repo_root hash_or_b =
-    let is_hex str =
-      OpamStd.String.fold_left (fun hex ch ->
-          hex && match ch with
-          | '0'..'9' | 'A'..'F' | 'a'..'f' -> true
-          | _ -> false
-        ) true str
-    in
     (* get the hash of the branch *)
     let hash =
       git repo_root ["branch"] @@> fun r ->
@@ -288,13 +284,13 @@ module VCS : OpamVCS.VCS = struct
         if is_branch then
           git repo_root [ "rev-list"; hash_or_b; "-1" ] @@> fun r ->
           if OpamProcess.is_success r then
-            (match List.filter is_hex r.r_stdout with
+            (match List.filter OpamStd.String.is_hex r.r_stdout with
              | [hash] ->
                Done (Some hash)
              | _ -> Done None)
           else Done None
         else
-        if is_hex hash_or_b then
+        if OpamStd.String.is_hex hash_or_b then
           Done (Some hash_or_b)
         else Done None
       else
